/**
 * EasyBeans
 * Copyright (C) 2006-2012 Bull S.A.S.
 * Copyright (C) 2013 Peergreen S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id:PersistenceXmlFileAnalyzer.java 1537 2007-07-08 15:31:22Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.persistence.basic;

import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.List;

import javax.persistence.spi.PersistenceProvider;
import javax.persistence.spi.PersistenceProviderResolverHolder;
import javax.transaction.TransactionManager;

import org.ow2.easybeans.persistence.EZBPersistenceUnitManager;
import org.ow2.easybeans.persistence.EZBPersistenceXmlAnalyzer;
import org.ow2.easybeans.persistence.PersistenceXmlAnalyzerException;
import org.ow2.easybeans.persistence.manager.PersistenceUnitManager;
import org.ow2.easybeans.persistence.util.ClassUtils;
import org.ow2.easybeans.persistence.xml.JPersistenceUnitInfo;
import org.ow2.easybeans.persistence.xml.JPersistenceUnitInfoException;
import org.ow2.easybeans.persistence.xml.JPersistenceUnitInfoHelper;
import org.ow2.util.archive.api.ArchiveException;
import org.ow2.util.archive.api.IArchive;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Class used to analyze a given archive (by analyzing the persistence.xml file
 * if any).
 * @author Florent Benoit
 */
public class BasicPersistenceXmlAnalyzer implements EZBPersistenceXmlAnalyzer {

    /**
     * Directory where persistence.xml file should be.
     */
    private static final String DIRECTORY_PERSISTENCE_XML_FILE = "META-INF";

    /**
     * Another possibility for persistence.xml file location in case of WARs.
     */
    private static final String WEB_DIRECTORY_PERSISTENCE_XML_FILE = "WEB-INF/classes/META-INF";

    /**
     * Name of the persistence.xml file.
     */
    private static final String PERSISTENCE_XML_FILE = "persistence.xml";

    /**
     * Name of the orm.xml file.
     */
    private static final String ORM_XML_FILE = "orm.xml";

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(BasicPersistenceXmlAnalyzer.class);

    /**
     * Transaction manager.
     */
    private TransactionManager transactionManager;


    /**
     * Detects and analyze the META-INF/persistence.xml file.
     * @param archive the file to analyze (or directory) in order to find a
     *        persistence.xml file.
     * @param classloader the classloader used to load the archive
     * @return a a persistence unit manager
     * @throws PersistenceXmlAnalyzerException if detection or analyze
     *         fails.
     */
    @Override
    public EZBPersistenceUnitManager analyzePersistenceXmlFile(final IArchive archive, final ClassLoader classLoader)
            throws PersistenceXmlAnalyzerException {
        URL persistenceXmlURL = null;

        try {
            persistenceXmlURL = archive.getResource(DIRECTORY_PERSISTENCE_XML_FILE + '/' + PERSISTENCE_XML_FILE);
        } catch (ArchiveException e) {
            throw new PersistenceXmlAnalyzerException("Cannot check if entry '" + DIRECTORY_PERSISTENCE_XML_FILE
                    + '/' + PERSISTENCE_XML_FILE + "' is present on the file '" + archive.getName() + "'.", e);
        }

        URL ormXmlURL = null;
        try {
            ormXmlURL = archive.getResource(DIRECTORY_PERSISTENCE_XML_FILE + '/' + ORM_XML_FILE);
        } catch (ArchiveException e) {
            throw new PersistenceXmlAnalyzerException("Cannot check if entry '" + DIRECTORY_PERSISTENCE_XML_FILE
                    + '/' + ORM_XML_FILE + "' is present on the file '" + archive.getName() + "'.", e);
        }

        if (persistenceXmlURL == null) {
            try {
                persistenceXmlURL = archive.getResource(WEB_DIRECTORY_PERSISTENCE_XML_FILE + '/' + PERSISTENCE_XML_FILE);
            } catch (ArchiveException e) {
                throw new PersistenceXmlAnalyzerException("Cannot check if entry '" + WEB_DIRECTORY_PERSISTENCE_XML_FILE
                        + '/' + PERSISTENCE_XML_FILE + "' is present on the file '" + archive.getName() + "'.", e);
            }
        }


        try {
            return analyzePersistenceXmlFile(archive.getURL().toURI(), persistenceXmlURL.toURI(), ormXmlURL.toURI(), classLoader);
        } catch (URISyntaxException | ArchiveException e) {
            throw new PersistenceXmlAnalyzerException("Cannot check if there is persistence in archive '" + archive.getName() + "'.", e);
        }
    }


    /**
     * Analyze the given persistence.xml file.
     * @param persistenceUnitRooUrl the root URL of the persistence unit
     * @param persistenceXMLURI the persistence.xml file to analyze.
     * @param ormXMLURI the orm.xml file to analyze (optional).
     * @param classloader the classloader used to load the archive
     * @return a a persistence unit manager
     * @throws PersistenceXmlAnalyzerException if analyze fails.
     */
    @Override
    public EZBPersistenceUnitManager analyzePersistenceXmlFile(final URI persistenceUnitRooUrl, final URI persistenceXMLURI, final URI ormXMLURI, final ClassLoader classLoader)
            throws PersistenceXmlAnalyzerException {


        // Now, do the parsing and fill the structure.
        List<JPersistenceUnitInfo> persistenceUnitInfos = null;
        if (persistenceXMLURI != null) {
            try {
                persistenceUnitInfos = JPersistenceUnitInfoHelper.getPersistenceUnitInfoList(persistenceXMLURI.toURL());
            } catch (JPersistenceUnitInfoException e) {
                throw new PersistenceXmlAnalyzerException("Cannot parse the URL '" + persistenceXMLURI + "'.", e);
            } catch (MalformedURLException e) {
                throw new PersistenceXmlAnalyzerException("Cannot parse the URL '" + persistenceXMLURI + "'.", e);
            }
            JPersistenceUnitInfo[] persistenceUnitInfosTab = persistenceUnitInfos.toArray(new JPersistenceUnitInfo[persistenceUnitInfos.size()]);
            for (JPersistenceUnitInfo persistenceUnitInfo : persistenceUnitInfos) {
                try {
                    persistenceUnitInfo.setPersistenceUnitRootUrl(persistenceUnitRooUrl.toURL());
                } catch (MalformedURLException e) {
                    throw new PersistenceXmlAnalyzerException("Cannot parse the URL '" + persistenceUnitRooUrl + "'.", e);
                }

                // Add mapping file
                if (ormXMLURI != null) {
                    persistenceUnitInfo.addMappingFileName(DIRECTORY_PERSISTENCE_XML_FILE + '/' + ORM_XML_FILE);
                }
            }
            logger.debug("Size of Persistence Unit Infos: ''{0}''", persistenceUnitInfosTab.length);
            return loadPersistenceProvider(persistenceUnitInfosTab, classLoader);
        }
        // nothing found, return nothing(null)
        return null;
    }

    /**
     * Load in the classloader the persistence provider class.
     * @param persistenceUnitInfos a list of persistence unit information.
     * @param classLoader the classloader used to load the persistence provider
     *        class.
     * @return A persistence unit manager (which can manage the persistence contexts).
     * @throws PersistenceXmlAnalyzerException if detection or analyze
     *         fails.
     */
    protected PersistenceUnitManager loadPersistenceProvider(final JPersistenceUnitInfo[] persistenceUnitInfos,
                                                                 final ClassLoader classLoader)
                                                                 throws PersistenceXmlAnalyzerException {

        for (JPersistenceUnitInfo persistenceUnitInfo : persistenceUnitInfos) {
            // sets the classloader
            persistenceUnitInfo.setClassLoader(classLoader);


            // Gets the JPA 2.0 list of persistence providers
            List<PersistenceProvider> persistenceProviders = PersistenceProviderResolverHolder.getPersistenceProviderResolver().getPersistenceProviders();


            // No provider defined by the user
            if (persistenceUnitInfo.getPersistenceProviderClassName() == null
                    && persistenceUnitInfo.getPersistenceProvider() == null) {
                if (persistenceProviders != null && persistenceProviders.size() > 0) {
                    PersistenceProvider provider = persistenceProviders.get(0);
                    persistenceUnitInfo.setPersistenceProvider(provider);
                    persistenceUnitInfo.setPersistenceProviderClassName(provider.getClass().getName());
                } else {
                    throw new PersistenceXmlAnalyzerException("No Persistence provider has been set");
                }
            } else if (persistenceUnitInfo.getPersistenceProviderClassName() != null && persistenceUnitInfo.getPersistenceProvider() == null) {

                // Provider defined by the user, check if there is a matching persistence provider
                if (persistenceProviders != null && persistenceProviders.size() > 0) {
                    int i = 0;
                    boolean matching = false;
                    while (matching == false && i < persistenceProviders.size()) {
                        PersistenceProvider persistenceProvider = persistenceProviders.get(i);
                        if (persistenceProvider.getClass().getName().equals(persistenceUnitInfo.getPersistenceProviderClassName())) {
                            // match
                            matching = true;
                            persistenceUnitInfo.setPersistenceProvider(persistenceProvider);
                        }
                        i++;
                    }
                }
            }


            // Instantiate only if this was not already done
            if (persistenceUnitInfo.getPersistenceProvider() == null) {

                // instantiate persistence provider
                Class<?> persistenceProviderClass;
                try {
                    persistenceProviderClass = ClassUtils.forName(persistenceUnitInfo
                            .getPersistenceProviderClassName(), BasicPersistenceXmlAnalyzer.class);
                } catch (ClassNotFoundException e) {
                    throw new PersistenceXmlAnalyzerException("Cannot load the persistence provider class '"
                            + persistenceUnitInfo.getPersistenceProviderClassName() + "'.");
                }
                PersistenceProvider persistenceProvider;
                try {
                    persistenceProvider = (PersistenceProvider) persistenceProviderClass.newInstance();
                } catch (InstantiationException e) {
                    throw new PersistenceXmlAnalyzerException("Cannot instantiate the persistence provider class '"
                            + persistenceUnitInfo.getPersistenceProviderClassName() + "'.", e);
                } catch (IllegalAccessException e) {
                    throw new PersistenceXmlAnalyzerException("Cannot instantiate the persistence provider class '"
                            + persistenceUnitInfo.getPersistenceProviderClassName() + "'.", e);
                }

                // Set persistence provider
                persistenceUnitInfo.setPersistenceProvider(persistenceProvider);
            }
        }
        // create persistence unit manager
        return new PersistenceUnitManager(persistenceUnitInfos, transactionManager);
    }

    /**
     * Sets the transaction manager.
     * @param transactionManager the transaction manager
     */
    public void setTransactionManager(TransactionManager transactionManager) {
        this.transactionManager = transactionManager;
    }

}
